/* boot-anim-start
 * Copyright 2009 Bobby Powers
 * Copyright (C) 2009 One Laptop per Child
 * Released under the GPLv2+, see COPYING for details
 *
 * Performs initial setup for graphical boot, and waits for a signal to
 * indicate that either the animation has completed or that the user has moved
 * to a different virtual terminal.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include "utils.h"

/* frame 25 for testing, for now */
static const char *DEFAULT_IMAGE = "/usr/share/boot-anim/frame00.565";
static const char *DELTA_FILE = "/usr/share/boot-anim/deltas";

static unsigned short *fb;
static unsigned int *state;
static volatile int got_signal = 0;

static void
init_diffs (unsigned short *fb)
{
  int deltas;
  size_t delta_size;

  state[FRAME_NUM] = 0;
  /* give us some room incase there is state information I'm forgetting */
  state[OFFSET] = FB_SIZE + 16;

  deltas = open (DELTA_FILE, O_RDONLY);
  if (deltas < 0)
  {
    write (1, "Couldn't open delta file.\n", 26);
    return;
  };

  delta_size = read (deltas, &fb[state[OFFSET]], EXTRA_SPACE);
  //printf ("Read in %dk worth of deltas, at %x\n", delta_size/1024, state[OFFSET]);

  fb[state[OFFSET] + delta_size/2 + 0] = 0x3544;
  fb[state[OFFSET] + delta_size/2 + 1] = 0x3536;

  close (deltas);
}

static void
vt_change_handler (int signo)
{
  got_signal = signo;
}

int
main (void)
{
  int ret = 1;
  int fb0, tty1, tty2;
  int console = -1;
  sigset_t mask, mask_old;
  struct sigaction sa;
  struct vt_mode newmode = {
    .mode = VT_PROCESS,
    .waitv = 0,
    .relsig = SIGUSR2,
    .acqsig = 0,
    .frsig = 0,
  };

  fb0 = open ("/dev/fb0", O_RDWR);
  if (fb0 < 0)
  {
    write (1, "Couldn't open /dev/fb0 for boot animation.\n", 44);
    return 1;
  }

  fb = (unsigned short *)mmap (NULL,
                               FB_SIZE_B + EXTRA_SPACE,
                               PROT_READ|PROT_WRITE,
                               MAP_SHARED,
                               fb0,
                               0);
  state = (unsigned int *)&fb[FB_SIZE];

  memset(&sa, 0, sizeof (sa));
  sa.sa_handler = vt_change_handler;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGUSR2, &sa, NULL);
  sigaction(SIGALRM, &sa, NULL);

  sigemptyset(&mask);
  sigaddset(&mask, SIGUSR2);
  sigaddset(&mask, SIGALRM);
  sigprocmask(SIG_BLOCK, &mask, &mask_old);

  /* redirect console output to tty1, and set terminal to graphics mode */
  /* step 1 - disable cursor on VT2 */
  tty2 = open ("/dev/tty2", O_WRONLY);
  if (tty2 < 0)
  {
    write (1, "Couldn't open /dev/tty2 to set mode.\n", 37);
    goto cleanup;
  }

  ioctl (tty2, VT_SETMODE, &newmode);
  ioctl (tty2, KDSETMODE, KD_GRAPHICS);

  /* change to vt2 */
  console = open ("/dev/console", O_RDWR);
  if (console < 0)
  {
    write (1, "Couldn't open /dev/console to switch VTs.\n", 42);
    goto cleanup;
  }
  chvt_in_userspace (console, 2);

  /* redirect console messages to vt1, even though cscott's notes
   * indicate this doesn't work. */
  tty1 = open ("/dev/tty1", O_RDWR);
  if (tty1 < 0)
  {
    write (1, "Couldn't open /dev/tty1 for redirection.\n", 41);
    goto cleanup;
  }
  ioctl (tty1, TIOCCONS);
  close (tty1);

  /* write the default image, which we then will paint image diffs over. */
  state[RUNNING] = 1;
  write_image (fb, DEFAULT_IMAGE);
  init_diffs (fb);

  dcon_freeze (0);

  /* tell parent that we're ready */
  kill (getppid(), SIGUSR1);

  /* at this point, we just have to sit around waiting for a VT change to
   * happen, at which point a SIGUSR2 will cause us to clean up and exit.
   * this 3 minute timeout is here in case something goes wrong. */
  alarm (180);
  sigsuspend (&mask_old);
  alarm (0);
  sigprocmask (SIG_SETMASK, &mask_old, NULL);

  state[RUNNING] = 0;
  if (got_signal == SIGUSR2) {
    /* VT switch OK */
    ioctl (tty2, VT_RELDISP, 1);
    ret = 0;
  }

cleanup:
  munmap (fb, FB_SIZE_B);
  close (fb0);

  if (tty2 >= 0) {
    /* back to VT_AUTO and text mode */
    ioctl (tty2, KDSETMODE, KD_TEXT);
    memset (&newmode, 0, sizeof (newmode));
    ioctl (tty2, VT_SETMODE, &newmode);
    close (tty2);
  }

  if (console >= 0) {
    /* disable console redirection */
	/* FIXME: disabled because this doesn't have the desired effects.
	 * if we abort the splash early during boot, everything is fine. but if
	 * we abort while starting the services in the runlevel, we lose all
	 * console output from the point where the splash is turned off. ??? */
    //ioctl (console, TIOCCONS);
    close (console);
  }

  return ret;
}

