/* boot-anim-start
 * Copyright 2009 Bobby Powers
 * Copyright (C) 2009 One Laptop per Child
 * Released under the GPLv2+, see COPYING for details
 *
 * starts our graphical boot, but only if the DCON is frozen
 */
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>

#include "utils.h"

static int calc_steps (unsigned short *fb);

static void
sighandler (int signo)
{
  /* SIGUSR1 means that the child has finished setting everything up */
  _exit (0);
}

int
main (int argc, char *argv[])
{
  sigset_t mask;
	struct sigaction sa;
  int fb_file = 0;
  unsigned short *fb = NULL;
  int r;

  /* we only animate for prettyboot, which freezes the DCON first */
  if (!dcon_frozen ()) return 0;

  fb_file = open ("/dev/fb0", O_RDWR);
  if (fb_file < 0)
  {
    write (1, "Couldn't open /dev/fb0 for boot animation.\n", 44);
    return 1;
  }

  fb = (unsigned short *)mmap (NULL,
                               FB_SIZE_B + EXTRA_SPACE,
                               PROT_READ|PROT_WRITE,
                               MAP_SHARED,
                               fb_file,
                               0);

  /* this script seems to get called twice (at the beginning of each
   * runlevel), and messes up our nicely choreographed boot animation
   * if we restart the animation the second time it was called. */
  if (boot_anim_started (&fb[FB_SIZE]))
    goto err;

  set_started (&fb[FB_SIZE]);

  if (calc_steps (fb) < 0)
    goto err;

  munmap (fb, FB_SIZE_B);
  close (fb_file);

  memset(&sa, 0, sizeof (sa));
  sa.sa_handler = sighandler;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGUSR1, &sa, NULL);

  sigemptyset(&mask);
  sigaddset(&mask, SIGUSR1);
  sigprocmask(SIG_BLOCK, &mask, NULL);

  r = fork();
  if (r < 0) {
    return 1;
  } else if (r == 0) {
    /* child */
    /* unblock SIGUSR1 */
    sigprocmask(SIG_UNBLOCK, &mask, NULL);
    execl("/usr/sbin/boot-anim-helper", "boot-anim-helper", NULL);
    _exit (1);
  }

  /* parent */

  /* wait for SIGUSR1 from the child, to indicate that we're ready to
   * continue with boot. use 30-second timeout just in case something goes
   * wrong. */
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
  sleep(30);

  return 1; /* the normal exit path is through _exit(0) inside sighandler */

err:
  munmap (fb, FB_SIZE_B);
  close (fb_file);
  return 1;
}

/* Calculate the number of steps that we should render the animation for.
 * This is based on the 4 initial RC state updates that come in
 * (RCkernelparam, RChostname, RCmountfs and RCswap) plus the number of services
 * that get started in runlevel 5. */
static int
calc_steps (unsigned short *fb)
{
  DIR *dir = opendir("/etc/rc.d/rc5.d");
  struct dirent *ent;
  unsigned int *state = (unsigned int *)&fb[FB_SIZE];

  /* initialize at 2, calculated from:
   * 4 initial RC state updates that come in addition to services
   * minus 2 ignored services that come in runlevel 5
   * (0-boot-anim-start and z-boot-anim-stop) */
  int ctr = 2;

  if (!dir) {
    write (1, "Couldn't open rc5.d directory.\n", 31);
    return -1;
  }

  while ((ent = readdir(dir)))
    if (ent->d_name[0] == 'S')
      ctr++;

  state[STEP] = 0;
  state[TOTAL_STEPS] = ctr;
  return 0;
}


