/* boot-anim-client
 * copyright 2009 Bobby Powers, released under the GPLv2+
 * see COPYING for details
 *
 * follows the calling convention of rhgb-client
 */
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "utils.h"

static
unsigned short *draw_frame (unsigned short *fb, unsigned short *info)
{
  /* skip the "D565" header */
  info = &info[2];

  while (!is_header (info))
  {
    unsigned short line = info[LINE];
    unsigned short first_col = info[FIRSTCOL];
    unsigned short len = info[LEN];

    memcpy (&fb[FB_WIDTH*line + first_col], &info[DATA], len*2);

    info = &info[DATA + len];
  }

  return info;
}

int
main (int argc, char *argv[])
{
  int fb_file = 0;
  unsigned short *fb = NULL;
  unsigned short *info;
  unsigned int *state;
  float progress_fraction;
  float frame_position;
  unsigned int frames_to_draw = 1;
  unsigned int show_frame;
  unsigned int next_to_draw;

  /* if we're not called with 'start' as an arg, quit */
  if (argc < 2) return 0;

  /* support plymouth --hide-splash */
  if (strcmp (argv[1], "--hide-splash") == 0) {
    execl ("/usr/sbin/boot-anim-stop", "boot-anim-stop", NULL);
    return 1;
  }

  /* only animate if we're getting an update argument */
  if (strlen (argv[1]) < 8 || strncmp (argv[1], "--update", 8)) return 0;

  /* copied from cscott's original boot-anim-client */
  if (!strcmp (argv[1], "--update=0-boot-anim-start") ||
      !strcmp (argv[1], "--update=z-boot-anim-stop") ||
      !strcmp (argv[1], "--update=ul-warning") ||
      !strcmp (argv[1], "--update=vprocunhide")) return 0;

  /* update animation frame here, if we meet the requirements. */
  fb_file = open ("/dev/fb0", O_RDWR);
  if (fb_file < 0)
  {
    write (1, "Couldn't open /dev/fb0 for boot animation.\n", 44);
    return 0;
  }

  fb = (unsigned short *)mmap (NULL,
                               FB_SIZE_B + EXTRA_SPACE,
                               PROT_READ|PROT_WRITE,
                               MAP_SHARED,
                               fb_file,
                               0);

  state = (unsigned int *)&fb[FB_SIZE];

  /* make sure that we're animating */
  if (!boot_anim_started (state) || !state[RUNNING])
    goto cleanup;

  if (state[FRAME_NUM] >= LAST_FRAME) goto cleanup;

  /* jump to the beginning of the next diff */
  info = &fb[state[OFFSET]];

  /* looking for 'D565' */
  if (!is_header (info)) goto cleanup;

  /* based on the number of times we expected to be invoked (the number of
   * "steps"), we have to decide if we want to move onto the next frame in
   * the animation or not. if there are a low number of total steps, we might
   * even need to render more than one frame.*/

  state[STEP]++;
  if (state[TOTAL_STEPS] == 0)
    goto cleanup; /* should never happen */

  /* on a scale of 0 to 1, how far through the animation should we be? */
  progress_fraction = (float) state[STEP] / (float) state[TOTAL_STEPS];

  /* how does that fraction relate to a position in our sequence of frames? */
  frame_position = progress_fraction * (LAST_FRAME + 1);

  /* OK, which frame should we be showing on screen at this step? */
  show_frame = (int) (frame_position + 0.5); /* cheap man's round */
  show_frame--; /* because it's zero-based */

  /* the frame we want to draw might already be on-screen */
  if (show_frame <= state[FRAME_NUM])
    goto cleanup;

  if (show_frame > LAST_FRAME)
    show_frame = LAST_FRAME; /* should never happen */

  /* the next frame in the sequence from the one that is on-screen */
  next_to_draw = state[FRAME_NUM] + 1;

  /* we might have to skip a frame or two, however we can't just skip them
   * because these deltas only render correctly if they are all rendered,
   * in series. so we just end up drawing multiple frames on this invocation. */
  if (show_frame > next_to_draw)
    frames_to_draw += show_frame - next_to_draw;

  while (frames_to_draw-- > 0)
    info = draw_frame(fb, info);

  state[FRAME_NUM] = show_frame;
  state[OFFSET] = info - fb;

  /* disabled for now: we need a new solution for dcon unfreezing */
#if 0
  if (state[FRAME_NUM] == LAST_FRAME) dcon_freeze (1);
#endif

cleanup:
  munmap (fb, FB_SIZE_B);
  close (fb_file);

  return 0;
}

