/* utils.c
 * copyright 2009 Bobby Powers, released under the GPLv2+
 * see COPYING for details
 *
 * common useful and interesting functions
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <linux/kd.h>
#include <linux/vt.h>

#include "utils.h"

static const char *DCON_FREEZE = "/sys/devices/platform/dcon/freeze";


/* boot-anim-start seems to get called twice, perhaps first when we
 * enter runlevel 1, then again when we enter runlevel 5.  this is to
 * check if its the first run or not, so that we don't erase all the
 * progress of our boot animation.
 * This function indicates whether we were started at all during this
 * bootup. It doesn't account for the fact that we might have finished
 * or have been stopped early. */
inline int
boot_anim_started (void *state)
{
  unsigned int *s = (unsigned int *)state;

  return s[COOKIE] == 0xbeeeeeef;
}


inline void
set_started (void *state)
{
  unsigned int *s = (unsigned int *)state;

  s[COOKIE] = 0xbeeeeeef;
}


inline void
clear_started (void *state)
{
  unsigned int *s = (unsigned int *)state;

  s[COOKIE] = 0;
}


/* test to see if the location pointed to by p represents a
 * diff file header, which is the ascii "D565" */
inline int is_header (void *p)
{
  char *pc = (char *)p;
  return (pc[0] == 0x44 && pc[1] == 0x35 &&
          pc[2] == 0x36 && pc[3] == 0x35);
}


int
dcon_frozen ()
{
  int dcon_freeze;
  char value;

  dcon_freeze = open (DCON_FREEZE, O_RDONLY);

  /* if we can't open the dcon, just report unfrozen */
  if (dcon_freeze < 0) return 0;

  read (dcon_freeze, &value, sizeof (value));
  close (dcon_freeze);

  /* 0 is unfrozen, 1 is frozen */
  return value == '0' ? 0 : 1;
}


void
dcon_freeze (int freeze)
{
  int dcon_freeze;
  char value;

  value = freeze ? '1' : '0';

  dcon_freeze = open (DCON_FREEZE, O_WRONLY);

  if (dcon_freeze < 0)
  {
    write (1, "Couldn't open DCON for writing.\n", 32);
  }

  write (dcon_freeze, &value, sizeof (value));
  close (dcon_freeze);
}

void
write_image (unsigned short *fb, const char* file_name)
{
  int infile;
  unsigned char header[12];

  int imagewidth, imageheight;

  infile = open (file_name, O_RDONLY);

  if (infile < 0)
  {
    write (1, "Couldn't open image for reading.\n", 33);
    return;
  }

  if (read (infile, header, sizeof (header)) < 0)
  {
    write (1, "Couldn't read image header.\n", 28);
    goto cleanup;
  }

  /* from WMB's calcdelta.c file */
  imagewidth  = (header[3] << 24) + (header[2] << 16) +
                (header[1] << 8) + header[0];
  imageheight = (header[7] << 24) + (header[6] << 16) +
                (header[5] << 8) + header[4];

  if (imagewidth != FB_WIDTH || imageheight != FB_HEIGHT)
  {
    write (1, "Image size different from framebuffer's.\n", 41);
    goto cleanup;
  }

  /* read the input file directly into the framebuffer */
  read (infile, fb, FB_SIZE_B);

cleanup:
  close (infile);
}

/* change active terminal and verify the switch without using VT_WAITACTIVE
 * which is inherently racy (imagine 2 chvt processes running at the same time).
 * Will aggressively try changing terminal several times.
 * returns 1 on success, 0 if it could never verify the switch, and negative
 * on error. */
int
chvt_in_userspace (int fd, int num)
{
#define USER_WAIT_SLEEP_US		100000
#define USER_WAIT_MAX_ITERATIONS	30
  int i;

  for (i = 0; i < USER_WAIT_MAX_ITERATIONS; i++) {
    struct vt_stat vtstat;
    if (ioctl (fd, VT_GETSTATE, &vtstat))
      return -1;

    if (vtstat.v_active == num)
      return 1;

    if (ioctl (fd, VT_ACTIVATE, num))
      return -1;
    usleep(USER_WAIT_SLEEP_US);
  }
  return 0;
}
